///////////////////////////////////////////////////////////////////////////////
// ControllerGL.cpp
// ================
// Derived Controller class for OpenGL window
// It is the controller of OpenGL rendering window. It initializes DC and RC,
// when WM_CREATE called, then, start new thread for OpenGL rendering loop.
//
//  AUTHOR: Song Ho Ahn (song.ahn@gmail.com)
// CREATED: 2006-07-09
// UPDATED: 2006-07-09
//
//	EXTENSION:
//	AUTHOR: Lukas Chripko (luxas98@gmail.com)
//  UPDATED: 2011-05-23
///////////////////////////////////////////////////////////////////////////////

#include <process.h>                                // for _beginthreadex()
#include "ControllerGL.h"
#include "Window.h"
#include "Log.h"
#include <string>
using namespace Win;

void MarshalString ( System::String ^ s, std::string& os ) {
	using namespace System::Runtime::InteropServices;
	const char* chars = 
      (const char*)(Marshal::StringToHGlobalAnsi(s)).ToPointer();
	os = chars;
	Marshal::FreeHGlobal(System::IntPtr((void*)chars));
}
void MarshalString ( System::String ^ s, std::wstring& os ) {
	using namespace System::Runtime::InteropServices;
	const wchar_t* chars = 
      (const wchar_t*)(Marshal::StringToHGlobalUni(s)).ToPointer();
	os = chars;
	Marshal::FreeHGlobal(System::IntPtr((void*)chars));
}


///////////////////////////////////////////////////////////////////////////////
// default contructor
///////////////////////////////////////////////////////////////////////////////
ControllerGL::ControllerGL(ModelGL* model, ViewGL* view, ControllerFormGL* formCtrl) : model(model), view(view),
                                                           threadHandle(0), threadId(0), loopFlag(false), formCtrl(formCtrl)
{
	pom = 0;
}



///////////////////////////////////////////////////////////////////////////////
// handle WM_CLOSE
///////////////////////////////////////////////////////////////////////////////
int ControllerGL::close()
{
    loopFlag = false;
    ::WaitForSingleObject(threadHandle, INFINITE);   // wait for rendering thread is terminated
    // close OpenGL Rendering context
    view->closeContext(handle);

    ::DestroyWindow(handle);
    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// handle WM_CREATE
///////////////////////////////////////////////////////////////////////////////
int ControllerGL::create()
{
    // create a OpenGL rendering context
    if(!view->createContext(handle, 32, 24, 8))
    {
        Win::log(L"[ERROR] Failed to create OpenGL rendering context from ControllerGL::create().");
        return -1;
    }

    // create a thread for OpenGL rendering
    // The params of _beginthreadex() are security, stackSize, functionPtr, argPtr, initFlag, threadId.
    threadHandle = (HANDLE)_beginthreadex(0, 0, (unsigned (__stdcall *)(void *))threadFunction, this, 0, &threadId);
    if(threadHandle)
    {
        loopFlag = true;
        Win::log(L"Created a rendering thread for OpenGL.");
    }
    else
    {
        Win::log(L"[ERROR] Failed to create rendering thread from ControllerGL::create().");
    }

    return 0;
}



///////////////////////////////////////////////////////////////////////////////
// handle WM_PAINT
///////////////////////////////////////////////////////////////////////////////
int ControllerGL::paint()
{
    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// handle WM_COMMAND
///////////////////////////////////////////////////////////////////////////////
int ControllerGL::command(int id, int cmd, LPARAM msg)
{
    return 0;
}



///////////////////////////////////////////////////////////////////////////////
// route to worker thread
// The function prototype is:
// unsigned int (__stdcall *)(void *)
///////////////////////////////////////////////////////////////////////////////
void ControllerGL::threadFunction(void* param)
{
    ((ControllerGL*)param)->runThread();
}



///////////////////////////////////////////////////////////////////////////////
// rendering thread
// initialize OpenGL states and start rendering loop
///////////////////////////////////////////////////////////////////////////////
void ControllerGL::runThread()
{
    // set the current RC in this thread
    ::wglMakeCurrent(view->getDC(), view->getRC());
    // initialize OpenGL states
    model->init();
    Win::log(L"Initialized OpenGL states.");

    // cofigure projection matrix
    RECT rect;
    ::GetClientRect(handle, &rect);
    model->setViewport(rect.right, rect.bottom);
	//model->setViewport(0,0);
    Win::log(L"Initialized OpenGL viewport and projection matrix.");

    // rendering loop
    Win::log(L"Entering OpenGL rendering thread...");
    while(loopFlag)
    {
        ::Sleep(1);                     // yield to other processes or threads
        model->draw();
        view->swapBuffers();
    }

    // terminate rendering thread
    ::wglMakeCurrent(0, 0);             // unset RC
    ::CloseHandle(threadHandle);
    Win::log(L"Exit OpenGL rendering thread.");
}



///////////////////////////////////////////////////////////////////////////////
// handle Left mouse down
///////////////////////////////////////////////////////////////////////////////
int ControllerGL::lButtonDown(WPARAM state, int x, int y)
{    // update mouse position
	//System::Windows::Forms::MessageBox::Show("lButtonDown, netusim ale kto zavola tuto picnu funkciu"); //vola sa ale nechapem odkial lebo procedure.cpp to nejak nehandluje, takze predsa len to vola procedure.cpp
	model->setMousePosition(x, y);
    if(state == MK_LBUTTON)
    {
		//System::Windows::Forms::MessageBox::Show("LEFT BUTTON MOUSE : " + x + " " + y);
        model->setMouseLeft(true);
		model->activateDimension();
    }
	
    return 0;
}



///////////////////////////////////////////////////////////////////////////////
// handle Left mouse up
///////////////////////////////////////////////////////////////////////////////
int ControllerGL::lButtonUp(WPARAM state, int x, int y)
{
    // update mouse position
		
		model->deactivateDimension();
		model->setMouseLeft(false);
		model->setMousePosition(x, y);
    return 0;
}



///////////////////////////////////////////////////////////////////////////////
// handle reft mouse down
///////////////////////////////////////////////////////////////////////////////
int ControllerGL::rButtonDown(WPARAM state, int x, int y)
{
    // update mouse position
    if(state == MK_RBUTTON)
    {
		//System::Windows::Forms::MessageBox::Show("RIGHT BUTTON MOUSE : " + x + " " + y);
        model->setMouseRight(true);
		model->setMouseRecPos(x,y);
    }
	model->setMousePosition(x, y);
    return 0;
}



///////////////////////////////////////////////////////////////////////////////
// handle reft mouse up
///////////////////////////////////////////////////////////////////////////////
int ControllerGL::rButtonUp(WPARAM state, int x, int y)
{
    // update mouse position
    model->setMousePosition(x, y);
    model->setMouseRight(false);
    return 0;
}
///////////////////////////////////////////////////////////////////////////////
// handle WM_MOUSEMOVE
///////////////////////////////////////////////////////////////////////////////
int ControllerGL::mouseMove(WPARAM state, int x, int y)
{
    if(state == MK_LBUTTON)
    {
		model->setMousePosition(x,y);
    }
    if(state == MK_RBUTTON)
    {
        model->setMousePosition(x,y);
		model->activateRecords();
    }
	model->moveActiveDimension();
	//pom++;
    return 0;
}
int ControllerGL::size(int w, int h, WPARAM wParam){
	model->setViewport(w,h);
	return 0;
}
int ControllerGL::resize(int x, int y){
	//System::Windows::Forms::MessageBox::Show("VOLAJU NIEKEDY RESIZE NA CTRLGL?");
	//::SetWindowPos(getHandle(),0,0,0,x,y,SWP_NOZORDER);
	//Win::Window mainWin = (Win::Window)getHandle();
	//System::Windows::Forms::MessageBox::Show(((Win::Window)getHandle())
	//::SetWindowPos(getHandle(),0,0,0,pom,pom,SWP_NOZORDER);
	::SetWindowPos(getHandle(),0,0,0,x,y,SWP_NOZORDER);
	model->setViewport(x, y);
	return 0;
}
int ControllerGL::loadFile(){
	model->opened = false;
	//model->myHandler->~Handler();
	//std::vector<record*>* myRecords;
	StreamReader^ myStream;
	String^ line;
	array<String^>^ pom;
	OpenFileDialog^ fileDialog = gcnew OpenFileDialog;
	//fileDialog->Filter = "okc files (*.okc)|*.okc|All files (*.*)|*.*";
	fileDialog->Filter = "okc files (*.okc)|*.okc";

	//Ak otvorime subor dojde k prerusenia aktualneho vlakna jeho zrusenie a vytvorenie noveho vlakna pre nas subor
	if(fileDialog->ShowDialog() == DialogResult::OK){
		loopFlag = false;
		::WaitForSingleObject(threadHandle, INFINITE);
		view->closeContext(handle);

		// create a OpenGL rendering context
		if(!view->createContext(handle, 32, 24, 8))
		{
			Win::log(L"[ERROR] Failed to create OpenGL rendering context from ControllerGL::create().");
			return -1;
		}

		// create a thread for OpenGL rendering
		// The params of _beginthreadex() are security, stackSize, functionPtr, argPtr, initFlag, threadId.
		threadHandle = (HANDLE)_beginthreadex(0, 0, (unsigned (__stdcall *)(void *))threadFunction, this, 0, &threadId);
		if(threadHandle)
		{
			loopFlag = true;
			Win::log(L"Created a rendering thread for OpenGL.");
		}
		else
		{
			Win::log(L"[ERROR] Failed to create rendering thread from ControllerGL::create().");
		}
	//VLAKNO HOTOVE, OPENGL BEZI
	//START NACITAVANIA
			myStream = gcnew StreamReader(fileDialog->FileName);
			line = myStream->ReadLine();
			int i = 1;
			int test = line->Length;
			
			pom = line->Replace('\t',' ')->Split(' ');
			//model->myHandler->Dimensions->clear();
			//model->myHandler->records->clear();
			if(pom->Length < 3){
		//myInfo = AllocateDynamicArray<double>(System::Int32::Parse(pom[0]->ToString()),System::Int32::Parse(pom[1]->ToString()));
			model->dim_size = System::Int32::Parse(pom[0]->ToString());
			model->record_size = System::Int32::Parse(pom[1]->ToString());
	}

	//VYPOCET VZDIALENOSTI DIMENZI
	model->vzd_dim = model->windowWidth / model->dim_size;
	//INICIALIZACIA POLI PRE DIMENSION
	model->max_dim = new float[model->dim_size];
	model->min_dim = new float[model->dim_size];
	model->name_dim = new std::string[model->dim_size];
	model->dim_number = new int[model->dim_size];
	model->dim_pos_x = new int[model->dim_size];
	model->dim_pos_x_init = new int[model->dim_size];

	model->threshold = new double[model->dim_size];
	int faktorial = 1;
	for(int i = 1; i <= model->dim_size; i++){
		faktorial = faktorial * i;
	}
	//System::Windows::Forms::MessageBox::Show(model->dim_size.ToString());
	//System::Windows::Forms::MessageBox::Show(faktorial.ToString());
	model->outliner_numbers = new int[faktorial];
	model->permutation = (int**)malloc(faktorial* sizeof(int*));
	for(int i = 0; i < faktorial ; i++){
		model->permutation[i] = (int*)malloc(model->dim_size* sizeof(int*));
	}	
	for(int i = 0; i < model->dim_size; i++){
		std::string marshal_string;
		MarshalString(myStream->ReadLine(),marshal_string);
		model->name_dim[i] = marshal_string;
		model->threshold[i] = 0;
		//model->addDimension(myStream->ReadLine());		
	}
		for(int i = 0; i < model->dim_size; i++){
			pom = myStream->ReadLine()->Replace('\t',' ')->Split(' ');
		
			pom[0] = pom[0]->Replace('.',',');
			pom[1] = pom[1]->Replace('.',',');
			model->min_dim[i] = System::Double::Parse(pom[0]->ToString());
			model->max_dim[i] = System::Double::Parse(pom[1]->ToString());
			model->dim_pos_x[i] = (25 + i*model->vzd_dim);
			model->dim_pos_x_init[i] = 25 + i*model->vzd_dim;
		}
		//int poc = 0;
		model->records = (double**)malloc(model->record_size * sizeof(double*));
		for(int i = 0; i < model->record_size; i++){
			model->records[i] = (double*)malloc(model->dim_size * sizeof(double));
		}
		model->outliner_record = (bool**)malloc(model->record_size* sizeof(bool*));
		for(int i = 0; i < model->record_size; i++){
			model->outliner_record[i] = (bool*)malloc((model->dim_size-1)* sizeof(bool));
		}
		for(int i = 0; i < model->record_size;i++)
			for(int j = 0; j < model->dim_size-1 ; j++)
				model->outliner_record[i][j] = false;
		model->poc_pridanych = 0;
		formCtrl->resetListInfo();

		model->mean = new double[model->dim_size];


		while( (line = myStream->ReadLine()) != nullptr){
			line = line->Replace('.',',');
			line = line->Replace('\t',' ');

			model->addRecord(line);
			//formCtrl->addListInfo(line);
		}
		model->active_records = new bool[model->record_size];
		for(int i = 0; i < model->record_size; i++){
			model->active_records[i] = false;
		}
		myStream->Close();
		model->opened = true;
	}
	//UNIT myMsg;
	//::PostMessage(getHandle(),UNIT(),0,0);
	return 0;
}